
import aiohttp
import asyncio
import io
import pandas as pd
import zipfile

from fast_binance.archieve_files import AbstractFile

from fast_binance.utils import (
    chunked_iterable
)

class OfflineFileFetcher:
    def __init__(self):
        self._worker = 250

    async def fetch_files(self, files:list[AbstractFile]):
        res = []
        async with aiohttp.ClientSession() as session:
            for file_chunk in chunked_iterable(files, self._worker):
                res.extend(await self._fetch_chunk(session, file_chunk))
        return res

    async def _fetch_chunk(self, session, files:list[AbstractFile]):
        tasks = []
        for file in files:
            task = asyncio.ensure_future(self._fetch_file(session, file))
            tasks.append(task)
        return await asyncio.gather(*tasks, return_exceptions=True)

    async def _fetch_file(self, session, file):
        '''
        downloads zip file and extract .csv file and add column information
        returns pandas dataframe
        '''
        print(file.source)
        async with session.get(file.source) as resp:
            assert resp.status == 200
            data = await resp.read()
            with zipfile.ZipFile(io.BytesIO(data)) as archive:
                fname = archive.namelist()[0]
                data = pd.read_csv(archive.open(fname), dtype=object, header=None)
                data = file.prepare_df(data)
        return data
    
    def download(self, files):
        loop = asyncio.get_event_loop()
        res = loop.run_until_complete(self.fetch_files(files))
        return res



from itertools import islice

def chunked_iterable(iterable, chunk_size):
    """
    slices the given iterable into chunks with chunk_size size.
    
    """
    it = iter(iterable)
    while True:
        chunk = list(islice(it, chunk_size))
        if not chunk:
            break
        yield chunk



import asyncio

from binance import AsyncClient, Client

import pandas as pd

from fast_binance.utils import chunked_iterable

DEFAULT_WORKER_COUNT = 30

CANDLE_COLUMNS  = ['open_time', 'open', 'high', 'low', 'close', 'volume', 
                    'close_time', 'quote_volume', 'count', 
                    'taker_buy_volume', 'taker_buy_quote_volume', 'ignore']

OI_STATS_COLUMNS = ['symbol', 'sumOpenInterest', 'sumOpenInterestValue', 'timestamp']

  
def convert_price_data_types(raw_prices:pd.DataFrame, convert_types=True):
    """ converts non-processed price data into useful DataFrame format.

    Args:
        raw_prices (pd.DataFrame): non-indexed, no types defined DataFrame

    Returns:
        pd.DataFrame: DateTimeIndex'ed, type defined DataFrame
    """
    raw_prices.columns = CANDLE_COLUMNS
    raw_prices['open_time'] = pd.to_datetime(raw_prices['open_time'], unit='ms')
    # raw_prices = raw_prices.drop(['ignore', 'close_time', 'taker_buy_volume'], axis=1)
    if convert_types:
        raw_prices = raw_prices.astype({'open': 'float64', 
                                        'close': 'float64', 
                                        'high':'float64', 
                                        'low':'float64',
                                        'volume':'float64',
                                        'count':'int64',
                                        'quote_volume':'float64',
                                        'taker_buy_volume':'float64',
                                        'taker_buy_quote_volume':'float64'})
    raw_prices = raw_prices.set_index('open_time')
    return raw_prices


def convert_oi_data(oi_data:pd.DataFrame, convert_types=True):
    oi_data.columns = OI_STATS_COLUMNS
    if convert_types:
        oi_data = oi_data.astype({
            'timestamp': 'int64', 
            'sumOpenInterest': 'float64', 
            'sumOpenInterestValue':'float64',
    })
    oi_data['timestamp'] = pd.to_datetime(oi_data['timestamp'], unit='ms')
    oi_data = oi_data.set_index('timestamp')
    return oi_data


class HistoricalFundingRate:
    def __init__(self):
        self._client = Client()

    def download(self):
        response = self._client.futures_funding_rate(limit=1000)
        funding_rate = pd.DataFrame(response)
        funding_rate['fundingTime'] = pd.to_datetime(funding_rate['fundingTime'], unit='ms')
        funding_rate = funding_rate.set_index('symbol')
        return funding_rate
    

class MarkPrice:
    def __init__(self):
        self._client = Client()

    def download(self):
        response = self._client.futures_mark_price()
        mark_price = pd.DataFrame(response)
        mark_price['nextFundingTime'] = pd.to_datetime(mark_price['nextFundingTime'], unit='ms')
        mark_price['time'] = pd.to_datetime(mark_price['time'], unit='ms')
        mark_price = mark_price.set_index('symbol')
        return mark_price

class PriceResult:
    def __init__(self, symbols, interval, start, end):
        pass
    
    def save(self):
        pass

    def load(self):
        pass

class MultiplexFetcher:
    def __init__(self, worket_count = DEFAULT_WORKER_COUNT):
        self._worker = worket_count # Parallel async request count
        self._async_client = None

    async def _fetch_symbols(self, symbols, **kwargs):
        self._async_client = await AsyncClient.create()
        results = []

        try:
            for symbols_in_chunk in chunked_iterable(symbols, self._worker):
                res = await self._fetch_chunk(symbols_in_chunk, **kwargs)
                results.extend(res)
        finally:
            await self._async_client.close_connection()
        
        result_dict = { result[0]: result[1] for result in results }
        return result_dict
    
    async def _fetch_chunk(self, symbols, **kwargs):
        """
        creates a Future object for every symbol in the symbol_chunk.
        it awaits until all the Future objects returns.
        
        """
        tasks = []
        for symbol in symbols:
            task = asyncio.ensure_future(self._fetch_symbol(symbol, **kwargs))
            tasks.append(task)
        return await asyncio.gather(*tasks, return_exceptions=True)

    async def _fetch_symbol(self, symbol,  **kwargs):
        raise NotImplementedError
    
    async def _get_function(self, symbol, **kwargs):
        raise NotImplementedError
    
    async def async_download(self, symbols, **kwargs):
        return await self._fetch_symbols(symbols, **kwargs)
    
    def download(self, symbols, **kwargs):
        try:
            loop = asyncio.get_event_loop()
        except RuntimeError as e:
            if str(e).startswith('There is no current event loop in thread'):
                loop = asyncio.new_event_loop()
                asyncio.set_event_loop(loop)
            else:
                raise
        res = loop.run_until_complete(self._fetch_symbols(symbols, **kwargs))
        return res
    

class OnlinePriceFetcher(MultiplexFetcher):
    def __init__(self, market, worket_count = DEFAULT_WORKER_COUNT):
        super().__init__(worket_count)
        self._market = market
        
    async def _fetch_symbol(self, symbol, **kwargs):
        """
        bottom-level function that sends HTTP request to Binance and awaits on response
        """
        try:
            price_data_raw = await self._get_function(symbol,  **kwargs)
            price_df = pd.DataFrame(price_data_raw)
            price_df = convert_price_data_types(price_df)
        except Exception as e:
            print(f'An error occured while fetching {symbol}. Error: {e}. {kwargs}')
            # raise Exception(f'An error occured while fetching {symbol}. Error: {e}')
            return symbol, str(e)
        return symbol, price_df


    async def _get_function(self, symbol, **kwargs):
        # TODO symbol unavailable hatasi geldiginde onu sonuca eklememek lazim
        if self._market == 'spot':
            return await self._async_client.get_historical_klines(symbol,  **kwargs)
        elif self._market == 'futures':
            return await self._async_client.futures_historical_klines(symbol,  **kwargs)
        else:
            raise ValueError(f"market type is invalid : {self._market}")

class OpenInterestFetcher(MultiplexFetcher):
    async def _get_function(self, **kwargs):
        # TODO symbol unavailable hatasi geldiginde onu sonuca eklememek lazim
        return await self._async_client.futures_open_interest_hist(**kwargs)
    
    async def _fetch_symbol(self, symbol, **kwargs):
        """
        bottom-level function that sends HTTP request to Binance and awaits on response
        """
        try:
            kwargs['symbol'] = symbol
            oi_raw = await self._get_function(**kwargs)
            oi_df = pd.DataFrame(oi_raw)
            oi_df = convert_oi_data(oi_df)
        except Exception as e:
            print(f'An error occured while fetching {symbol}. Error: {e}. {kwargs}')
            # raise Exception(f'An error occured while fetching {symbol}. Error: {e}')
            return symbol, str(e)
        return symbol, oi_df
        


import pandas as pd


class AbstractFile:
    _base_url = 'https://data.binance.vision/data'
    @property
    def source(self):
        raise NotImplementedError('source() method should be implemented in derived classes.')

    def prepare_df(self, df):
        raise NotImplementedError('prepare_df() method should be implemented in derived classes.')


class KlinesFile(AbstractFile):
    columns = ['open_time', 'open', 'high', 'low', 'close', 'volume', 'close_time', 
               'quote_volume', 'count', 'taker_buy_volume', 'taker_buy_quote_volume', 'ignore']
     
    def __init__(self, market, symbol, period, date, span):
        self.market = 'spot' if market == 'spot' else 'futures/um'
        self.symbol = symbol
        self.period = period
        self.date = date
        self.span = span
        self.name = f'{self.symbol}-{self.period}-{self.date}'

class MetricsFile:
    def __init__(self, symbol, date):
        self.symbol = symbol
        self.date = date
        self._base_url = 'https://data.binance.vision/data/futures/um'
        
    @property
    def source(self):
        return  f'{self._base_url}/daily/metrics/{self.symbol}/{self.symbol}-metrics-{self.date}.zip'


class PriceKlinesFile(KlinesFile):
    type_mapping = {'open_time':int, 'open':float, 'high':float, 'low':float, 'close':float,
                    'volume':float, 'quote_volume':float, 'taker_buy_volume':float }

    def __init__(self, market, symbol, period, date, span='daily'):
        super().__init__(market, symbol, period, date, span)

    @property
    def source(self):
        '''
        https://data.binance.vision/data/futures/um/daily/klines/ACHUSDT/15m/ACHUSDT-15m-2024-02-07.zip
        '''
        return f'{self._base_url}/{self.market}/{self.span}/klines/{self.symbol}/{self.period}/{self.name}.zip'

    def prepare_df(self, df:pd.DataFrame):
        try:
            df.columns = self.columns
            df = df.astype(self.type_mapping)
            df['open_time'] = pd.to_datetime(df['open_time'], unit='ms')
            df.set_index('open_time', inplace=True)
        except Exception as e:
            print(e)
        finally:
            return df

class FuturesUsdtMetricsFile(AbstractFile):
    def __init__(self, symbol, date):
        self.binance = MetricsFile(symbol, date)

    def prepare_df(self, df):
        try:
            #TODO
            pass
        except Exception as e:
            print(e)
        finally:
            return df

class PremiumIndexKlinesFile(KlinesFile):
    type_mapping = {'open_time':int, 'open':float, 'high':float, 'low':float, 'close':float }

    def __init__(self, symbol, period, date, span):
        # Can we find span by looking at date?
        super().__init__('futures', symbol, period, date, span)
        
    
    def prepare_df(self, df:pd.DataFrame):
        try:
            df = df.astype(self.type_mapping)
            df['open_time'] = pd.to_datetime(df['open_time'], unit='ms')
            df.set_index('open_time', inplace=True)
            df.drop(columns=['close_time','volume','taker_buy_volume', 
                             'quote_volume', 'taker_buy_quote_volume', 
                             'ignore'], inplace=True)
        except Exception as e:
            print(e)
        finally:
            return df
        
    @property
    def source(self):
        '''
        https://data.binance.vision/data/futures/um/monthly/premiumIndexKlines/TRBUSDT/1m/TRBUSDT-1m-2023-09.zip
        '''
        return f'{self._base_url}/futures/um/{self.span}/premiumIndexKlines/' \
            f'{self.symbol}/{self.period}/{self.symbol}-{self.period}-{self.date}.zip'

class MarkPriceKlines:
    # TODO
    pass


class PremiumIndexFile:
    def __init__(self, symbol, date, span, period):
        self.symbol = symbol
        self.date = date
        self.period = period
        self._span = span
        self._base_url = 'https://data.binance.vision/data/futures/um'
    
    @property
    def source(self):
        '''
        https://data.binance.vision/data/futures/um/monthly/premiumIndexKlines/TRBUSDT/1m/TRBUSDT-1m-2023-09.zip
        '''
        return f'{self._base_url}/{self._span}/premiumIndexKlines/' \
            f'{self.symbol}/{self.period}/{self.symbol}-{self.period}-{self.date}.zip'

class AggregatedTradesFile(AbstractFile):
    def __init__(self, symbol, date, span, market):
        self.symbol = symbol
        self.date = date
        self.span = span
        self.market = 'spot' if market == 'spot' else 'futures/um'
        self.name = f'{self.symbol}-aggTrades-{self.date}' # ACAUSDT-aggTrades-2024-02-07
        self.columns = ['aggTradeID', 'price', 'quantity', 'ftId', 'ltId', 'time', 'is_buyer_maker', 'ig']  if market == 'spot' \
                        else ['aggTradeID', 'price', 'quantity', 'ftId', 'ltId', 'time', 'is_buyer_maker']

    @property
    def source(self):
        '''
        https://data.binance.vision/data/spot/daily/aggTrades/ACAUSDT/ACAUSDT-aggTrades-2024-02-07.zip
        https://data.binance.vision/data/futures/um/daily/aggTrades/ACAUSDT/ACAUSDT-aggTrades-2024-02-07.zip
        '''
        return f'{self._base_url}/{self.market}/{self.span}/aggTrades/{self.symbol}/{self.name}.zip'
    
    @property
    def local(self):
        pass

    def prepare_df(self, df):
        print('IM HERE 1', self.columns)
        df.columns = self.columns
        df = df.astype({"time":int})
        print('IM HERE 1b')
        df['time'] = pd.to_datetime(df['time'], unit='ms')
        print('IM HERE 2')
        df = df.drop(columns=['ftId', 'ltId', 'ig'])
        df = df.set_index(df['time']).sort_index()
        print('IM HERE 3')
        return df



__version__ = "1.1.1"

from fast_binance.online_fetcher import (
  OnlinePriceFetcher,
  OpenInterestFetcher
)

from fast_binance.offline_fetcher import OfflineFileFetcher

from fast_binance.archieve_files import(
    PriceKlinesFile,
    FuturesUsdtMetricsFile,
    MarkPriceKlines,
    PremiumIndexKlinesFile,
    AggregatedTradesFile
)


#!/usr/bin/env python
from setuptools import setup
import codecs
import os
import re


with codecs.open(
        os.path.join(
            os.path.abspath(os.path.dirname(__file__)),
            'fast_binance',
            '__init__.py'
        ), 'r', 'latin1') as fp:
    try:
        version = re.findall(r'^__version__ = "([^"]+)"\r?$', fp.read(), re.M)[0]
    except IndexError:
        raise RuntimeError('Unable to determine version.')

with open("README.rst", "r") as fh:
    long_description = fh.read()

setup(
    name='fast-binance',
    version=version,
    packages=['fast_binance'],
    description='Asyncful wrapper around python-binance module and data.binance.vision .',
    long_description=long_description,
    long_description_content_type="text/x-rst",
    url='https://github.com/mkaanerkoc/fast-binance',
    author='Kaan Erkoc',
    license='MIT',
    author_email='mkaanerkoc@gmail.com',
    install_requires=[
        'pandas', 'python-binance'
    ],
    keywords='binance exchange rest api bitcoin ethereum btc eth neo',
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
)